# Copyright Notice:
# Copyright 2016 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfishtool/blob/master/LICENSE.md

# redfishtool: Chassis.py
#
# contains Chassis subCommands and access functions
#
# Class RfSystemsMain
#  - functions init, displayUsage, displayHelp, displayOperations,
#  - runOperation - Chassis command table, dispatch of operation eg get, reset
#  - ChassisMain - called from redfishMain, enforce legal option combinations,
#    and call runOperation to run Chassis operations (sub-sub-command)
#
# Class RfChassisOperations
#  All of the Chassis sub-command operations eg: get, setIndicatorLed, etc
#  - hello - test cmd
#  - getCollection - return the Chassis collection
#  - get - get a member of a collection -or property of the member
#  - list - show of list of collection members and key idetifying properties
#      (Id, AssetTag, UriPath)
#  - patch - raw subcommand to patch a Chassis Member, with etag support
#  - setAssetTag -- patches the assetTag of Chassis instance w/ etag support
#  - setIndicatorLed --sets id LED to a specified value w/ etag support
#  - getPower - get Processors collection, processor instance, or all
#  - getThermal - get Ethernet collection, instance, all
#  - setPowerLimit - Set the PowerLimit for a member of the powerControl array
#  - getPowerReading [consumed] - get the powerControl array, or ConsumedWatts from pwrCntl[0]
#  - getLogService - get LogService collection, instance, all
#  - clearLog -- clears a specified log (not implemented yet)
#  - examples --prints some example apis
#
from   .redfishtoolTransport  import RfTransport
import requests
import json
import getopt
import re
import sys
from    .ServiceRoot import RfServiceRoot
from   urllib.parse import urljoin

class RfChassisMain():
    def __init__(self):
        # operation string and remaining args
        self.operation=None
        self.args=None
        self.argnum=0
        self.nonIdCommands=None

    def displayUsage(self,rft):
        if(rft.quiet): return(0)
        print("  Usage:")
        print("   {} [OPTNS]  Chassis  <operation> [<args>]  -- perform <operation> on the Chassis specified ".format(rft.program))

    def displayHelp(self,rft):
        self.displayUsage(rft)
        self.displayOperations(rft)
        print(" ")

    def displayOperations(self,rft):
        print("  <operations>:")
        print("     [collection]              -- get the main Chassis collection. (Default operation if no member specified)")
        print("     [get]                     -- get the Chassis object. (Default operation if collection member specified)")
        print("     list                      -- list information about the Chassis collection members(\"Id\", URI, and AssetTag)")
        print("     patch {A: B,C: D,...}     -- patch the json-formatted {prop: value...} data to the object")
        print("     setAssetTag <assetTag>    -- set the Chassis's asset tag ")
        print("     setIndicatorLed  <state>  -- set the indicator LED.  <state>=redfish defined values: Off, Lit, Blinking")
        print("     Power                     -- get the full Power resource under a specified Chassis instance.")
        print("     Thermal                   -- get the full Thermal resource under a specified Chassis instance.")
        print("")
        print("     getPowerReading [-i<indx>] [consumed]-- get powerControl resource w/ power capacity, PowerConsumed, and power limits")
        print("                                  if \"consumed\" keyword is added, then only current usage of powerControl[indx] is returned")
        print("                                  <indx> is the powerControl array index. default is 0.  normally, 0 is the only entry")
        print("     setPowerLimit [-i<indx>] <limit> [<exception> [<correctionTime>]] -- set powerLimit control properties")
        print("                               <limit>=null disables power limiting. <indx> is the powerControl array indx (dflt=0)")
        print("")
        print("     Logs [list]               -- get the Chassis \"LogServices\" collection , or list \"id\" and URI of members.")
        print("      Logs [IDOPTN]              --  get the member specified by IDOPTN: -i<id>, -m<prop>:<val>, -l<link>, -a #all")
        print("     clearLog   <id>           -- clears the log defined by <id>")
        print("     examples                  -- example commands with syntax")
        print("     hello                     -- Chassis hello -- debug command")
        return(0)



    def runOperation(self,rft):
        #  instantiate ChassisOperations class
        op=RfChassisOperations()
        
        #  dispatch table for each subcommand:   "cmdName": cmdClass.cmdFunction"
        operationTable = {
            "collection":                   op.getCollection,
            "get":                          op.get,
            "list":                         op.clist,
            "patch":                        op.patch,
            "setAssetTag":                  op.setAssetTag,
            "setIndicatorLed":              op.setIndicatorLed,
            "Power":                        op.getPower,
            "Thermal":                      op.getThermal,
            "setPowerLimit":                op.setPowerLimit,
            "getPowerReading":              op.getPowerReading,
            "Logs":                         op.getLogService,
            "-clearLog":                     op.clearLog,
            "hello":                        op.hello,
            "examples":                     op.examples
        }

        rft.printVerbose(5,"Chassis:runOperation: operation: {}".format(self.operation))
        rft.printVerbose(5,"Chassis:runOperation: args:  {}".format(self.args))
            
        if self.operation in operationTable:
            rft.printVerbose(5,"Chassis:runOperation: found Oper: {} in table. executing".format(rft.subcommand))
            rc,r,j,d=operationTable[self.operation](self, op, rft, cmdTop=True)
            return(rc,r,j,d)
        
        else: # invalid operation
            rft.printErr("Chassis: Invalid operation: {}".format(self.operation))
            return(2,None,False,None)
        


    def ChassisMain(self,rft,cmdTop=False):
        rft.printVerbose(4,"ChassisMain:  subcommand: {}".format(rft.subcommand))
        
        if( rft.help ):
            self.displayHelp(rft)
            return(0,None,False,None)
        
        # we will validate usage of -P and -a in action processing
        # actually, if a non 'get' action is specified, -P and -a are just ignored :)

        args=rft.subcommandArgv[0:]
        
        #if no args, then if no member Id was specified (with -I|-M|-1|-F) then assume it is a "collection" operation
        #            if a -IM1F was specified, then assume it is a "get" operation for that member
        if(  len(args) < 2 ):
            if( rft.IdOptnCount==0 ):
                self.operation="collection"
            else:
                self.operation="get"
            self.args= None
        else:
            self.operation=args[1]
            self.args = args[1:]        # now args points to the 1st argument
            self.argnum =len(self.args)
            
        rft.printVerbose(5,"Chassis: operation={}, args={}".format(self.operation,self.args))
                
        # check if the command requires a collection member target -I|-M|-L|-1|-F eg sysIdoptn
        nonIdCommands = ["collection", "list", "examples", "hello"]
        if( ( not self.operation in nonIdCommands ) and (rft.IdOptnCount==0) ):
            # default to --One if no Id option specified
            rft.oneOptn = True
            rft.IdOptnCount += 1
            
        # now execute the operation.
        rc,r,j,d = self.runOperation(rft)
                
        if(rc !=0 ):
            rft.printVerbose(5,"Chassis: operation returned with error: rc={}".format(rc))
            return(rc,r,False,None)
        
        #else, if here, the subcommand executed without error.  Return with 0 exit code
        rft.printVerbose(5,"Chassis: operation exited OK")
        return(rc,r,j,d)


#
# contains operations related to the Chassis subCommand
#
class RfChassisOperations():
    def __init__(self):
        self.chassisPath=None
        self.chassisCollectionDict=None


    def hello(self,sc,op,rft,cmdTop=False):
        rft.printVerbose(4,"in hello")
        rft.printVerbose(4,"   subcmd:{}, operation:{}, args:{}".format(rft.subcommand,sc.operation,sc.args))
        print("hello world from Chassis")
        return(0,None,False,None)
    
    def getCollection(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in getCollection".format(rft.subcommand,sc.operation))
        
        # 1st get serviceRoot
        svcRoot=RfServiceRoot()
        rc,r,j,d = svcRoot.getServiceRoot(rft)
        if( rc != 0 ):
            rft.printErr("getCollection: Error getting service root, aborting")
            return(rc,r,False,None)

        # get the link to the Systems collection
        # need to test we got good data
        if (("Chassis" in d) and ("@odata.id" in d["Chassis"])):
            systemsLink=d["Chassis"]["@odata.id"]
        else:
            rft.printErr("Error: service root does not have a Chassis link")
            return(4)
        
        rft.printVerbose(4,"Chassis:getCollection: link is: {}".format(systemsLink))


        # if a -a option was entered with "Chassis" or "Chassis collection" operation,
        # then return all members of the Chassis collection expanded
        if((cmdTop is True) and (rft.allOptn is True) ):
            collName="Chassis"
            rft.printVerbose(4,"Expand Chassis collection to return ALL Chassis collection members fully expanded in response")
            rc,r,j,d=rft.getAllCollectionMembers(rft, r.url, relPath=systemsLink)
            if(rc==0):
                rft.printVerbose(1," Get ALL {} Collection Members".format(collName,skip1=True, printV12=cmdTop))

        # otherwise, just return the collection
        # now read the /Chassis collection
        # use the returned url as the base url to read the Chassis collection
        else:
            if cmdTop is True:   prop=rft.prop
            rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=systemsLink, prop=prop)
            if(rc==0):
                rft.printVerbose(1," Chassis Collection:",skip1=True, printV12=cmdTop)
                
        return(rc,r,j,d)


    def get(self,sc,op,rft, cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
        
        # getCollection
        rc,r,j,d=op.getCollection(sc,op, rft)
        if( rc != 0):  return(rc,r,False,None)
        collUrl=r.url

        # search collection to find path to system
        sysPath,rc,r,j,d=rft.getPathBy(rft, r, d, prop)
        if( rc !=0 ):    #if a path was not found, its an error
            return(rc,r,j,d)
        
        rft.printVerbose(4,"ChassisOperations:get: got a path, now get entries")
        
        if cmdTop is True:   prop=rft.prop

        #if here, rc=0
        #if sysPath returned a response but we need to extract the property do it here
        if( (r is not None) and (prop is not None) ):
            rc,r,j,d=rft.getPropFromDict(rft,r,d,prop)
                
        # otherwise, we need to do a GET to get the Chassis, if -P show property, else show full response
        elif( r is None ):
            rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', collUrl, relPath=sysPath, prop=prop)

        if(rc==0):   rft.printVerbose(1," Chassis Resource:",skip1=True, printV12=cmdTop)
        return(rc,r,j,d)


    def clist(self,sc,op,rft, cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
        
        # getCollection
        collName="Chassis"
        rc,r,j,d=op.getCollection(sc,op, rft)
        if( rc != 0):  return(rc,r,False,None)
        #loop through the members and create the list sub-operation response
        rc,r,j,d=rft.listCollection(rft, r, d, prop="AssetTag")
        if(rc==0):
            rft.printVerbose(1," list {} Collection member info: Id, URI, AssetTag".format(collName,skip1=True, printV12=cmdTop))
        return(rc,r,j,d)


    def iterate_op(self, run_single, sc, op, rft, cmdTop=False, prop=None):
        # Wrapper method to handle issuing commands to a single chassis or to all chassis in the collection
        if rft.allOptn:
            # Issue command to all chassis in collection

            # get the list of chassis
            rc, r, j, d = op.clist(sc, op, rft, cmdTop=cmdTop, prop=prop)
            if rc != 0 or not j or d is None or not isinstance(d, dict) or 'Members' not in d:
                rft.printErr("Unable to get list of Chassis; return code = {}, list data = {}".format(rc, d))
                return rc, r, j, d

            # save existing rft options
            saved_allOptn = rft.allOptn
            saved_Link = rft.Link
            saved_gotIdOptn = rft.gotIdOptn
            saved_IdOptnCount = rft.IdOptnCount

            # set rft options to process a single item
            rft.allOptn = False
            rft.gotIdOptn = True
            rft.IdOptnCount = 1

            # iterate through chassis and run operation on each based on the link (@odata.id value)
            members = d.get('Members')
            rc, r, j, d = 8, None, False, None
            for member in members:
                if '@odata.id' in member:
                    link = member.get('@odata.id')
                    # set rft.Link to point to the target chassis
                    rft.Link = link
                    # perform the operation
                    rc, r, j, d = run_single(sc, op, rft, cmdTop=cmdTop, prop=prop)
                else:
                    rft.printErr("No '@odata.id' found in chassis member: {}".format(member))

            # restore existing rft options
            rft.allOptn = saved_allOptn
            rft.Link = saved_Link
            rft.gotIdOptn = saved_gotIdOptn
            rft.IdOptnCount = saved_IdOptnCount
            return rc, r, j, d
        else:
            # Issue command to single specified chassis
            return run_single(sc, op, rft, cmdTop=cmdTop, prop=prop)


    def patch_single(self,sc,op,rft,cmdTop=False, prop=None, patchData=None, r=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
        # verify we have got an argument which is the patch structure
        # its in form '{ "AssetTag": <val>, "IndicatorLed": <val> }'
        ##print("patchData: {}".format(patchData))
        if( len( sc.args) == 2):
            ##print("data:{}".format(sc.args[1]))
            try:
                patchData=json.loads(sc.args[1])
            except ValueError:
                rft.printErr("Patch: invalid Json input data:{}".format(sc.args[1]))
                return(5,None,False,None)
            ##print("patchData: {}".format(patchData))
        else:
            rft.printErr("Patch: error: invalid argument format")
            rft.printErr("     : args={}".format(sc.args))
            rft.printErr("     : expect: Systems patch \"{ <prop>: <value> }\"")
            return(4,None,False,None)

        # read the Chassis resource
        # this is used by the generic rft.patchResource() function to see if there is an etag in the response
        # if an etag is in response hdr, then we must include the etag on the patch, so this get is required
        # note: the format of etag header for redfish is:  ETag: W/"<string>" or "<string"
        rc,r,j,d=op.get(sc,op,rft)
        if( rc != 0):
            return(rc,r,False,None)

        # now call the generic patch function to send the patch
        rc,r,j,d=rft.patchResource(rft, r, patchData)

        if(rc==0):   rft.printVerbose(1," Chassis Patch:",skip1=True, printV12=cmdTop)
        return(rc,r,j,d)


    def patch(self, sc, op, rft, cmdTop=False, prop=None):
        return op.iterate_op(op.patch_single, sc, op, rft, cmdTop=cmdTop, prop=prop)


    def setAssetTag_single(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))

        propName="AssetTag"
        
        # get the asset tag from args
        if(len(sc.args) < 2 ):
            rft.printErr("Error, no assetTag value specified")
            rft.printErr("Syntax:  {} [options] Chassis setAssetTag \"string\" ".format(rft.program))
            return(8,None,False,None)
        assetTag=sc.args[1]
        patchData={propName: assetTag}
        
        # get the resource to verify it includes AssetTag,
        # the response will also be used by Patch() to check for etag 
        rc,r,j,d=op.get(sc,op,rft)
        if( rc != 0):  return(rc,r,False,None)
        if( not propName in d ):
            rft.printErr("Chassis resource does not have a {} property.".format(propName))
            return(8,r,False,None)

        rc,r,j,d=rft.patchResource(rft, r, patchData)
        if(rc==0):
            rft.printVerbose(1," Chassis setAssetTag:",skip1=True, printV12=cmdTop)
            assetTag={propName: d[propName]}
            return(rc,r,j,assetTag)         
        else: return(rc,r,False,None)


    def setAssetTag(self, sc, op, rft, cmdTop=False, prop=None):
        return op.iterate_op(op.setAssetTag_single, sc, op, rft, cmdTop=cmdTop, prop=prop)


    def setIndicatorLed_single(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))

        propName="IndicatorLED"
        
        # get the asset tag from args
        validLedStates=["Lit","Blinking","Off"]
        if(len(sc.args) < 2 ):
            rft.printErr("Error, no ledState value specified")
            rft.printErr("Syntax:  {} [options] Chassis setIndicatorLed <Lit|Off|Blinking>".format(rft.program))
            return(8,None,False,None)
        targLedState=sc.args[1]
        if not targLedState in validLedStates:
            rft.printErr("Error, invalid LED state specified")
            return(8,None,False,None)
        patchData={propName: targLedState}
        
        # get the resource to verify it includes IndicatorLED,
        # the response will also be used by Patch() to check for etag 
        rc,r,j,d=op.get(sc,op,rft)
        if( rc != 0):  return(rc,r,False,None)
        if( not propName in d ):
            rft.printErr("System resource does not have a {} property.".format(propName))
            return(8,r,False,None)

        rc,r,j,d=rft.patchResource(rft, r, patchData)
        if(rc==0):
            rft.printVerbose(1," Chassis setIndicatorLed:",skip1=True, printV12=cmdTop)
            ledState={"IndicatorLED": d["IndicatorLED"]}
            return(rc,r,j,ledState)
        else: return(rc,r,False,None)


    def setIndicatorLed(self, sc, op, rft, cmdTop=False, prop=None):
        return op.iterate_op(op.setIndicatorLed_single, sc, op, rft, cmdTop=cmdTop, prop=prop)


    def getPower(self,sc,op, rft, cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
        resName="Power"

        # get the Chassis resource first
        rc,r,j,d=op.get(sc, op, rft, cmdTop, resName)
        if( rc != 0):  return(rc,r,False,None)
        
        # get the link to the Power resource under Chassis
        if ((resName in d) and ("@odata.id" in d[resName])):
            resLink=d[resName]["@odata.id"]
        else:
            rft.printErr("Error: Chassis resource does not have a {} link".format(resName))
            return(6,None,False,None)
        
        if cmdTop is True:   prop=rft.prop

        rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=resLink, prop=prop)
        if(rc==0):
            rft.printVerbose(1," {} Resource ".format(resName,skip1=True, printV12=cmdTop))

        return(rc,r,j,d)



    def getThermal(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
        resName="Thermal"
        # get the Chassis resource first
        rc,r,j,d=op.get(sc, op, rft, cmdTop, resName)
        if( rc != 0):  return(rc,r,False,None)
        
        # get the link to the Thermal resource under Chassis
        if ((resName in d) and ("@odata.id" in d[resName])):
            resLink=d[resName]["@odata.id"]
        else:
            rft.printErr("Error: Chassis resource does not have a {} link".format(resName))
            return(6,None,False,None)
        
        if cmdTop is True:   prop=rft.prop
        
        rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=resLink, prop=prop)
        if(rc==0):
            rft.printVerbose(1," {} Resource ".format(resName,skip1=True, printV12=cmdTop))
            
        return(rc,r,j,d)



    def getPowerReading(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
                         
        if cmdTop is True:  prop=rft.prop

        # check if there is an arg for the operation
        if( sc.argnum > 1 ):
            if( sc.args[1] == 'consumed' ):
                prop="PowerConsumedWatts"
            else:
                rft.printErr("Error:  getPowerReading: invalid argument: {}".format(sc.args[1]))
                rft.printErr("Syntax: getPowerReading [-i<indx>] [consumed]")
                return(4,None,False,None)
            
        # get the Chassis Power resource
        rc,r,j,d=op.getPower(sc,op, rft)     
        if( rc != 0):  return(rc,r,False,None)
        
        # get the Power Control Object
        if ("PowerControl" in d ):
            powerControl=d["PowerControl"]
            powerControlMembers=len(powerControl)
        else:
            rft.printErr("Error: Chassis resource does not have a PowerControl resource")
            return(4,None,False,None)
        if(powerControlMembers == 0 ):
            rft.printErr("Error: Chassis PowerControl array is empty")
            return(4,None,False,None)

        # get the powerControl array index
        if( rft.IdLevel2 is None ):
            indx=0
        else:
            indxPattern="(^0$)|(^([1-9][0-9]{,2})$)"   #  <decimal>: 33...
            indxMatch=re.search(indxPattern,rft.IdLevel2)
            if( indxMatch ):
                indx=int(rft.IdLevel2)  # convert base10 to integer
                rft.printVerbose(4,"getPowerReading: Indx={}".format(indx))
            else:
                rft.printErr("Error: getPowerReading: invalid PowerControl index value: {}".format(rft.IdLevel2))
                rft.printErr("Chassis getPowerReading [current] [-i<indx>] --<indx> is integer 0 - 99 (default is 0)")
                         
            if( (indx+1) > powerControlMembers ):
                rft.printErr("Error: specified PowerControl index does not exist: indx={}, members={}".format(indx,powerControlMembers))
                return(4,None,False,None)
                        
        #if here, we have a valid indx
        if(prop is not None):
            if( prop in powerControl[indx] ):
                respDataVal=powerControl[indx][prop]
                respData={prop: respDataVal}
                rft.printVerbose(1," Get Current Power consumption (PowerConsumedWatts) of PowerControl[{}] resource".format(indx,skip1=True, printV12=cmdTop))
            else:
                rft.printErr("Error: Property {} not not returned in PowerControl[{}] resource".format(prop,indx))
                return(4,r,j,d)
        else:
            respData=powerControl[indx]  #return the full powerControl array
            rft.printVerbose(1," Chassis PowerControl[{}] array:".format(indx,skip1=True, printV12=cmdTop))
        
        return(rc,r,j,respData)




    #setPowerLimit [-i<indx>] <limit> [<exception> [<correctionTime>]] -- set powerLimit control properties")
    def setPowerLimit_single(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
                         
        if cmdTop is True:  prop=rft.prop

        # check if there is a required 2nd arg <limit> 
        if( sc.argnum > 1 ):
            limitPattern="(^null$)|(^0$)|(^([1-9][0-9]{,5})$)"   #  null | 0 | 8 | 33 |344 | 34333
            limitMatch=re.search(limitPattern,sc.args[1])  # sc.args[1]=<limit>
            if( limitMatch ):
                if(sc.args[1] == "null"):
                    limit=None
                else:
                    limit=int(sc.args[1])  # keep it a string
                rft.printVerbose(4,"setPowerLimit: limit={}".format(limit))
                powerLimitData={"LimitInWatts": limit }
            else:
                rft.printErr("Error: setPowerLimit: invalid <limit> value specified: {}".format(sc.args[1]))
                rft.printErr("Chassis setPowerLimit [-i<indx>] <limit> [<exception> [<correctionTime>]] --limit=null | 0-99999")
                return(8,None,False,None)
        else:
            rft.printErr("Error: setPowerLimit: no <limit> value specified: {}")
            rft.printErr("Chassis setPowerLimit [-i<indx>] <limit> [<exception> [<correctionTime>]] --limit=null | 0-99999")
            return(8,None,False,None)
        # check if there is an optional 3rd arg <exception>
        includeException=False
        validExceptionVals=("NoAction", "HardPowerOff", "LogEventOnly", "Oem")
        if (sc.argnum > 2 ):
            exceptionVal=sc.args[2] # sc.args[2]=<exceptionEnum>
            if( not exceptionVal in validExceptionVals ):
                rft.printErr("Error: setPowerLimit: invalid <limit> value specified: {}".format(sc.args[2]))
                rft.printErr("Chassis setPowerLimit [-i<indx>] <limit> [<exception> [<correctionTime>]] --limit=null | 0-99999")
                return(8,None,False,None)
            else:
                rft.printVerbose(4,"setPowerLimit: exception={}".format(exceptionVal))
                powerLimitData["LimitException"]=exceptionVal
                includeException=True
                         
        # check if there is an optional 4th arg <correctionTime> in miliseconds
        includeCorrectionTime=False
        if (sc.argnum > 3 ):
            correctionPattern="(^([1-9][0-9]{,6})$)"   #  1-999999 ms
            correctionMatch=re.search(correctionPattern,sc.args[3])  # sc.args[3]=<correctionInMs>
            if( correctionMatch ):
                correctionTime=int(sc.args[3])  # keep it a string
                rft.printVerbose(4,"setPowerLimit: correctionInMs={}".format(correctionTime))
                powerLimitData["CorrectionInMs"]=correctionTime 
                includeCorrectionTime=True
            else:
                rft.printErr("Error: setPowerLimit: invalid <correctionTime> value specified: {}".format(sc.args[3]))
                rft.printErr("Chassis setPowerLimit [-i<indx>] <limit> [<exception> [<correctionTime>]] --correctionTime=1-999999 ms")
                return(8,None,False,None)
                         
        # get the Chassis Power resource
        rc,r,j,d=op.getPower(sc,op, rft)     
        if( rc != 0):  return(rc,r,False,None)

        # get the Power Control Object
        if ("PowerControl" in d ):
            powerControl=d["PowerControl"]
            powerControlMembers=len(powerControl)
        else:
            rft.printErr("Error: Chassis resource does not have a PowerControl resource")
            return(4,None,False,None)
        if(powerControlMembers == 0 ):
            rft.printErr("Error: Chassis PowerControl array is empty")
            return(4,None,False,None)

        # get the powerControl array index
        if( rft.IdLevel2 is None ):
            indx=0
        else:
            indxPattern="(^0$)|(^([1-9][0-9]*)$)"   #  <decimal>: 33...
            indxMatch=re.search(indxPattern,rft.IdLevel2)
            if( indxMatch ):
                indx=int(rft.IdLevel2)  # convert base10 to integer
                rft.printVerbose(4,"setPowerLimit: Indx={}".format(indx))
            else:
                rft.printErr("Error: setPowerLimit: invalid PowerControl index: {}".format(rft.IdLevel2))
                rft.printErr("Chassis setPowerLimit [current] [-i<indx>] --<indx> is integer 0 - 99 (default is 0)")
                return(8,None,False,None)

            if( (indx+1) > powerControlMembers ):
                rft.printErr("Error: specified PowerControl index does not exist: indx={}, members={}".format(indx,powerControlMembers))
                return(8,None,False,None)

        # check that this PowerControl index member has the properties
        if( not "LimitInWatts" in powerControl[indx]["PowerLimit"] ):
            rft.printErr("Error: setPowerLimit: LimitInWatts property is not in rhost PowerControl[{}]".format(indx))
            return(8,None,False,None)
        if( ( includeException is True ) and ( not "LimitException" in powerControl[indx]["PowerLimit"] ) ):
            rft.printErr("Error: setPowerLimit: LimitException property is not in rhost PowerControl[{}]".format(indx))
            return(8,None,False,None)
        if( ( includeCorrectionTime is True ) and ( not "CorrectionInMs" in powerControl[indx]["PowerLimit"] ) ):
            rft.printErr("Error: setPowerLimit: CorrectionInMs property is not in rhost PowerControl[{}]".format(indx))
            return(8,None,False,None)

        # now build the final full patch data
        #patchData={"PowerControl"[indx]: { "PowerLimit":  powerLimitData } }
        powerLimitRes={ "PowerLimit": powerLimitData }
        patchData={ "PowerControl": [ {} ] }
        empty={}
        powerControlArray=list()
        for i in range (0,powerControlMembers):
            powerControlArray.append(empty)
        powerControlArray[indx]=powerLimitRes
        patchData["PowerControl"]=powerControlArray
        
        rft.printVerbose(4,"setPowerLimit: patchData: {}".format(json.dumps(patchData)))

        #call the generic patch command to send the patch.  This takes care of etag support
        rc,r,j,d=rft.patchResource(rft, r, patchData)

        if(rc==0):
            rft.printVerbose(1,"  SetPowerLimit [{}]:",indx, skip1=True, printV12=cmdTop)
            powerLimit={"PowerLimit": d["PowerControl"][indx]["PowerLimit"]}
            return(rc,r,j,powerLimit)
            
        else: return(rc,r,False,None)


    def setPowerLimit(self, sc, op, rft, cmdTop=False, prop=None):
        return op.iterate_op(op.setPowerLimit_single, sc, op, rft, cmdTop=cmdTop, prop=prop)

    
    def getLogService(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,":{}:{}: in operation".format(rft.subcommand,sc.operation))

        # get the Chassis resource
        rc,r,j,d=op.get(sc,op, rft)     
        if( rc != 0):  return(rc,r,False,None)
        
        collName="LogServices"
        # get the link to the LogServices collection
        if ((collName in d) and ("@odata.id" in d[collName])):
            logLink=d[collName]["@odata.id"]
        else:
            rft.printErr("Error: Chassis resource does not have a {} link".format(collName))
            return(6,None,False,None)

        if cmdTop is True:   prop=rft.prop

        # check if there is a list arg for the operation
        if( sc.argnum > 1 and sc.args[1] == 'list' ):
            #get the collection
            rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=logLink)
            #loop through the members and create the list sub-operation response
            rc,r,j,d=rft.listCollection(rft, r, d, prop="Name")
            if(rc==0):
                rft.printVerbose(1," list {} Collection member info: Id, URI, Name".format(collName,skip1=True, printV12=cmdTop))

        # else: check if no Log was specified.  If not, return the collection
        elif(rft.IdLevel2OptnCount==0):
            rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=logLink, prop=prop)
            if(rc==0):
                rft.printVerbose(1," {} Collection ".format(collName,skip1=True, printV12=cmdTop))

        # else:  check if the -a (all) option is set. If not, return the proc specific by -i or -m
        # search collection to find path using getPath2 
        elif( rft.allOptn is not True ):
            # get the LogServices collection
            rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=logLink, prop=prop)
            collUrl=r.url

            # now search for 2nd level resource and return
            path2,rc,r,j,d=rft.getLevel2ResourceById(rft,r,d)
            if(rc!=0):
                return(rc,r,j,d)
            # so rc=0
            #if sysPath returned a response but we need to extract the property do it here
            if( (r is not None) and (prop is not None) ):
                rc,r,j,d=rft.getPropFromDict(rft,r,d,prop)

            # otherwise, we need to do a GET to get the LogServices, if -P show property, else show full response
            elif( r is None ):
                rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', collUrl, relPath=path2, prop=prop)
                if(rc==0):
                    rft.printVerbose(1," {} Collection Member ".format(collName,skip1=True, printV12=cmdTop))

            # If '--Entries' specified, get "Entries" nav link and read it
            if rc == 0 and rft.gotEntriesOptn:
                if r is not None and j and isinstance(d, dict):
                    rft.printVerbose(1, 'getLogService: attempting to get Entries for Logs')
                    entries = d.get('Entries')
                    if entries is not None and isinstance(entries, dict):
                        entries_uri = entries.get('@odata.id')
                        if entries_uri is not None:
                            rc, r, j, d = rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url,
                                                                 relPath=entries_uri, prop=prop)
                        else:
                            rft.printErr('getLogService: @odata.id not found in "Entries" property')
                    else:
                        rft.printErr('getLogService: "Entries" property not found in JSON payload')
                else:
                    rft.printErr(
                        'Unable to fetch Entries property from previous response: response = {}, is_json = {}, type(json) = {}'
                        .format(r, j, type(d)))

        # else, client specified the -a option requesting ALL of the LogServices members
        # for logs, we will not support this.  Its too much data.
        else:
            rft.printErr("Error: -a option not supported for LogServices")
            return(6,None,False,None)
        
        return(rc,r,j,d)

    
    def clearLog(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
        return(8,None,False,None)


    def examples(self,sc,op,rft,cmdTop=False,prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
        print(" {} -r<ip> Chassis                          # shows the Chassis collection".format(rft.program))
        print(" {} -r<ip> Chassis list                     # lists Id, Uri, AssetTag for all Chassis".format(rft.program))
        print(" {} -r<ip> Chassis -I <id>                  # gets the Chassis with Id=<d>".format(rft.program))
        print(" {} -r<ip> Chassis -M AssetTag:12345        # gets the Chassis with AssetTag=12345".format(rft.program))
        print(" {} -r<ip> Chassis -L <sysUrl>              # gets the Chassis at URI=<systemUrl".format(rft.program))
        print(" {} -r<ip> Chassis -F                       # get the First Chassis returned (for debug)".format(rft.program))
        print(" {} -r<ip> Chassis -1                       # get the first Chassis and verify that there is only one system".format(rft.program))
        print(" {} -r<ip> Chassis -I <id> patch {{A: B,C: D,...}}     # patch the json-formatted {{prop: value...}} data to the object".format(rft.program))
        print(" {} -r<ip> Chassis -I <id> setAssetTag <assetTag>    # set the system's asset tag ".format(rft.program))
        print(" {} -r<ip> Chassis -I <id> setIndicatorLed  <state>  # set the indicator LED.  <state>=redfish defined values: Off, Lit, Blinking".format(rft.program))
        print(" {} -r<ip> Chassis -I<Id> Power             # get the full chassis Power resource".format(rft.program))
        print(" {} -r<ip> Chassis -I<Id> Thermal           # get the full chassis Thermal resource".format(rft.program))
        print(" {} -r<ip> Chassis -I<Id> getPowerReading[-i<indx> [consumed]   # get chassis/Power powerControl[<indx>] resource".format(rft.program))
        print("                                             # if optional \"consumed\" arg, then return only the PowerConsumedWatts prop")
        print(" {} -r<ip> Chassis -L<Url> setPowerLimit [-i<indx>] <limit> [<exception> [<correctionTime>]] # set power limit".format(rft.program))
        return(0,None,False,None)



    
'''
TODO:
1. clearlog not implemented

'''

    





# Copyright Notice:
# Copyright 2016 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfishtool/blob/master/LICENSE.md

# redfishtool: rawMain.py
#
# contains raw subCommands and access functions
#
# Class RfRawMain
#  - functions init, displayUsage, displayHelp, displayOperations,
#  - runOperation - raw subcommand table, dispatch of operations: get, patch, post..
#  - RfRawMain - called from redfishMain, enforce legal option combinations,
#    and call runOperation to run System operation (sub-sub-command)
#
# Class RfRawOperations
#  All of the Systems sub-command operations eg: Systems reset, setIndicatorLed, etc
#  - hello - test cmd
#  - httpGet    -- send GET method 
#  - httpPatch  -- send GET method
#  - httpPost   -- send GET method
#  - httpDelete -- send GET method
#  - httpHead   -- send GET method
#  - httpPut    -- send GET method  (not implemented in 0.9)
#  - examples --prints some example apis
#
from   .redfishtoolTransport  import RfTransport
import requests
import json
import getopt
import re
import sys
#from    .ServiceRoot import RfServiceRoot
from   urllib.parse import urljoin, urlparse, urlunparse

class RfRawMain():
    def __init__(self):
        # operation string and remaining args
        self.operation=None
        self.args=None
        self.argnum=0
        self.nonIdCommands=None

    def displayUsage(self,rft):
        if(rft.quiet): return(0)
        print("  Usage:")
        print("   {} [OPTNS] raw <method> <path> ".format(rft.program))
        print("")
        print("   {} raw -h        # for help".format(rft.program))
        print("   {} raw examples  #for example commands".format(rft.program))
        print("")
        print("  <method> is one of:  GET, PATCH, POST, DELETE, HEAD, PUT")
        print("  <path> is full URI path to a redfish resource--the full path following <ipaddr:port>, starting with forward slash /")
        print("")
        print("   Common OPTNS:")
        print("   -u <user>,   --user=<usernm>     -- username used for remote redfish authentication")
        print("   -p <passwd>, --password=<passwd> -- password used for remote redfish authentication")
        print("   -t <token>,  --token=<token>    - redfish auth session token-for sessions across multiple calls")
        print("")
        print("   -r <rhost>,  --rhost=<rhost>     -- remote redfish service hostname or IP:port")
        print("   -X <method>  --request=<method>  -- the http method to use. <method>={GET,PATCH,POST,DELETE,HEAD,PUT}. Default=GET")
        print("   -d <data>    --data=<data>       -- the http request \"data\" to send on PATCH,POST,or PUT requests")
        print("   -H <hdrs>, --Headers=<hdrs>      -- Specify the request header list--overrides defaults. Format \"{ A:B, C:D...}\" ")
        print("   -S <Secure>,  --Secure=<Secure>  -- When to use https: (Note: doesn't stop rhost from redirect http to https)")

        
    def displayHelp(self,rft):
        self.displayUsage(rft)
        self.displayOperations(rft)
        print("")
        
    def displayOperations(self,rft):
        print("  <operations / methods>:")
        print("     GET             -- HTTP GET method")
        print("     PATCH           -- HTTP PATCH method")
        print("     POST            -- HTTP POST method")
        print("     DELETE          -- HTTP DELETE method")
        print("     HEAD            -- HTTP HEAD method")
        print("     PUT             -- HTTP PUT method")
        print("   examples        -- example raw commands with syntax")
        print("   hello           -- raw hello -- debug command")
        return(0)


    def runOperation(self,rft):
        #  instantiate SystemsOperations class
        op=RfRawOperations()
        
        #  dispatch table for each subcommand:   "cmdName": cmdClass.cmdFunction"
        operationTable = {
            "GET":                  op.httpGet,
            "PATCH":                op.httpPatch,
            "POST":                 op.httpPost,
            "DELETE":               op.httpDelete,
            "HEAD":                 op.httpHead,
            "PUT":                  op.httpPut,
            "hello":                op.hello,
            "examples":             op.examples
        }

        rft.printVerbose(5,"raw: runOperation: operation: {}".format(self.operation))
        rft.printVerbose(5,"raw:runOperation:  args:  {}".format(self.args))
            
        if self.operation in operationTable:
            rft.printVerbose(5,"raw:runOperation: found Oper: {} in table. executing".format(rft.subcommand))
            rc,r,j,d=operationTable[self.operation](self, op, rft, cmdTop=True)
            return(rc,r,j,d)
        
        else: # invalid operation
            rft.printErr("raw: Invalid operation: {}".format(self.operation))
            return(2,None,False,None)
        


    def RawMain(self,rft,cmdTop=False):
        rft.printVerbose(4,"RawMain:  subcommand: {}".format(rft.subcommand))
        
        if( rft.help ):
            self.displayHelp(rft)
            return(2,None,False,None)
        
        args=rft.subcommandArgv[0:]
        
        # if 2 args, only operations are: hello and examples
        nonMethodOperations=("hello","examples")
        if( len(args)==2):
            nonMethodOp=args[1]
            if( not nonMethodOp in nonMethodOperations):
                rft.printErr("Syntax error: \"raw\" subcommands require 2 arguments\n")
                rft.printErr("")
                self.displayHelp(rft)
                return(2,None,False,None)
        # else all operations require two args <method> <path>
        elif(  len(args) < 3 ):
                    rft.printErr("Syntax error: \"raw\" subcommands require 2 arguments\n")
                    self.displayHelp(rft)
                    return(2,None,False,None)
        
        self.operation=args[1]      # for raw subcommand, this is the http method (eg GET)
        self.args = args[1:]        # now args points to the 1st argument
        self.argnum =len(self.args)
            
        rft.printVerbose(5,"raw: operation={}, args={}".format(self.operation,self.args))
                
           
        # now execute the operation.
        rc,r,j,d = self.runOperation(rft)
                
        if(rc !=0 ):
            rft.printVerbose(5,"raw: operation returned with error: rc={}".format(rc))
            return(rc,r,False,None)
        
        #else, if here, the subcommand executed without error.  Return with 0 exit code
        rft.printVerbose(5,"raw: operation exited OK")
        return(rc,r,j,d)


#
# contains operations related to the Systems subCommand
#
class RfRawOperations():
    def __init__(self):
        self.systemsPath=None
        self.systemsCollectionDict=None

    # function to return API type: authenticated or unauthenticated
    # authenticated APIs require authentication
    # unauthenticated APIs are /redfish, /redfish/v1, /redfish/v1/odata, /redfish/v1/$metadata
    def getApiType(self,rft,uri):
        unauthenticatedAPIs=("/redfish", "/redfish/v1", "/redfish/v1/", "/redfish/v1/odata", "/redfish/v1/$metadata")
        if not uri in unauthenticatedAPIs:
            return( rft.AUTHENTICATED_API )
        else:
            return( rft.UNAUTHENTICATED_API )

                             
    def hello(self,sc,op,rft,cmdTop=False):
        rft.printVerbose(4,"in hello")
        rft.printVerbose(4,"   subcmd:{}, operation:{}, args:{}".format(rft.subcommand,sc.operation,sc.args))
        print("hello world from raw subcommand")
        return(0,None,False,None)

    
    def httpGet(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in raw".format(rft.subcommand,sc.operation))
        
        # we verified that we had two args in RawMain(), so we can just read the <uri> arg here
        path=sc.args[1]
        method="GET"
        rft.printVerbose(4,"raw: GET: method:{} path:{}".format(method,path))

        apiType=self.getApiType(rft,path)   # UNAUTHENTICATED API or Authenticated API

        # calculate rootUrl--with correct scheme, root path, and rhost IP (w/o querying rhost)
        scheme=rft.getApiScheme(apiType)
        scheme_tuple=[scheme, rft.rhost, "/redfish", "","",""]
        rootUrl=urlunparse(scheme_tuple)  # so rootUrl="http[s]://<rhost>[:<port>]/redfish"
        
        if cmdTop is True:   prop=rft.prop
        jsonData=True
        if (path=="/redfish/v1/$metadata"): jsonData=False
        rc,r,j,d=rft.rftSendRecvRequest(apiType, method, rootUrl, relPath=path, prop=prop,jsonData=jsonData)
        if(rc==0):
            rft.printVerbose(1," raw GET:",skip1=True, printV12=cmdTop)  
            return(rc,r,j,d)
        else:
            rft.printErr("raw: Error getting response")
            return(rc,r,False,None)

    def httpHead(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in raw".format(rft.subcommand,sc.operation))
        
        # we verified that we had two args in RawMain(), so we can just read the <uri> arg here
        path=sc.args[1]
        method="HEAD"
        rft.printVerbose(4,"raw: HEAD: method:{} path:{}".format(method,path))

        apiType=self.getApiType(rft,path)   # UNAUTHENTICATED API or Authenticated API
        
        # calculate rootUrl--with correct scheme, root path, and rhost IP (w/o querying rhost)
        scheme=rft.getApiScheme(apiType)
        scheme_tuple=[scheme, rft.rhost, "/redfish", "","",""]
        rootUrl=urlunparse(scheme_tuple)  # so rootUrl="http[s]://<rhost>[:<port>]/redfish"
        
        rc,r,j,d=rft.rftSendRecvRequest(apiType, method, rootUrl, relPath=path)
        if(rc==0):
            rft.printVerbose(1," raw HEAD:",skip1=True, printV12=cmdTop)  
            return(rc,r,False,None)
        else:
            rft.printErr("raw: Error getting response")
            return(rc,r,False,None)


    

    def httpPatch(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in raw".format(rft.subcommand,sc.operation))

        # load patch data--verify its good json
        #  get the patchData from rft.requestData  readin on commandline via: -d <patchData>
        try:
            patchData=json.loads(rft.requestData)
        except ValueError:
            rft.printErr("Patch: invalid Json input data:{}".format(rft.requestData))
            return(5,None,False,None)
        ##print("patchData: {}".format(patchData))


        # we verified that we had two args in RawMain(), so we can just read the <uri> arg here
        path=sc.args[1]
        method="PATCH"
        rft.printVerbose(4,"raw: PATCH: method:{} path:{}".format(method,path))

        apiType=self.getApiType(rft,path)   # UNAUTHENTICATED API or Authenticated API

        # calculate rootUrl--with correct scheme, root path, and rhost IP (w/o querying rhost)
        scheme=rft.getApiScheme(apiType)
        scheme_tuple=[scheme, rft.rhost, "/redfish", "","",""]
        rootUrl=urlunparse(scheme_tuple)  # so rootUrl="http[s]://<rhost>[:<port>]/redfish"

        #read the resource--the generic patch command needs it to get the etag
        rc,r,j,d=rft.rftSendRecvRequest(apiType, "GET", rootUrl, relPath=path)
        if(rc!=0):
            rft.printErr("raw: Error getting resource prior to patching it, aborting")
            return(rc,r,False,None)

        # now call the generic patch function to send the patch
        rc,r,j,d=rft.patchResource(rft, r, patchData)

        if(rc==0):   rft.printVerbose(1," Systems Patch:",skip1=True, printV12=cmdTop)
        return(rc,r,j,d)


    def httpPost(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in raw".format(rft.subcommand,sc.operation))

        # load post data--verify its good json
        #  get the postData from rft.requestData  readin on commandline via: -d <patchData>
        try:
            postData=json.loads(rft.requestData)
        except ValueError:
            rft.printErr("Post: invalid Json input data:{}".format(rft.requestData))
            return(5,None,False,None)
        ##print("postData: {}".format(postData))

        # we verified that we had two args in RawMain(), so we can just read the <uri> arg here
        path=sc.args[1]
        method="POST"
        rft.printVerbose(4,"raw: POST: method:{} path:{}".format(method,path))

        apiType=self.getApiType(rft,path)   # UNAUTHENTICATED API or Authenticated API

        # calculate rootUrl--with correct scheme, root path, and rhost IP (w/o querying rhost)
        scheme=rft.getApiScheme(apiType)
        scheme_tuple=[scheme, rft.rhost, "/redfish", "","",""]
        rootUrl=urlunparse(scheme_tuple)  # so rootUrl="http[s]://<rhost>[:<port>]/redfish"

        #output the post data in json to send over the network   
        reqPostData=json.dumps(postData)
        #Post the data
        rc,r,j,d=rft.rftSendRecvRequest(apiType, method, rootUrl, relPath=path, reqData=reqPostData)
        if(rc!=0):
            rft.printErr("raw: Error sending POST to resource, aborting")
            return(rc,r,False,None)


        if(rc==0):   rft.printVerbose(1," raw POST:",skip1=True, printV12=cmdTop)
        return(rc,r,j,d)


    def httpPut(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in raw".format(rft.subcommand,sc.operation))

        # load put data--verify its good json
        #  get the postData from rft.requestData  readin on commandline via: -d <patchData>
        try:
            putData=json.loads(rft.requestData)
        except ValueError:
            rft.printErr("Put: invalid Json input data:{}".format(rft.requestData))
            return(5,None,False,None)
        ##print("postData: {}".format(postData))

        # we verified that we had two args in RawMain(), so we can just read the <uri> arg here
        path=sc.args[1]
        method="PUT"
        rft.printVerbose(4,"raw: POST: method:{} path:{}".format(method,path))

        apiType=self.getApiType(rft,path)   # UNAUTHENTICATED API or Authenticated API

        # calculate rootUrl--with correct scheme, root path, and rhost IP (w/o querying rhost)
        scheme=rft.getApiScheme(apiType)
        scheme_tuple=[scheme, rft.rhost, "/redfish", "","",""]
        rootUrl=urlunparse(scheme_tuple)  # so rootUrl="http[s]://<rhost>[:<port>]/redfish"

        #output the post data in json to send over the network   
        reqPutData=json.dumps(putData)
        #Put the data
        rc,r,j,d=rft.rftSendRecvRequest(apiType, method, rootUrl, relPath=path, reqData=reqPutData)
        if(rc!=0):
            rft.printErr("raw: Error sending PUT to resource, aborting")
            return(rc,r,False,None)


        if(rc==0):   rft.printVerbose(1," raw PUT:",skip1=True, printV12=cmdTop)
        return(rc,r,j,d)


    def httpDelete(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in raw".format(rft.subcommand,sc.operation))

        # we verified that we had two args in RawMain(), so we can just read the <uri> arg here
        path=sc.args[1]
        method="DELETE"
        rft.printVerbose(4,"raw: DELETE: method:{} path:{}".format(method,path))

        apiType=self.getApiType(rft,path)   # UNAUTHENTICATED API or Authenticated API

        # calculate rootUrl--with correct scheme, root path, and rhost IP (w/o querying rhost)
        scheme=rft.getApiScheme(apiType)
        scheme_tuple=[scheme, rft.rhost, "/redfish", "","",""]
        rootUrl=urlunparse(scheme_tuple)  # so rootUrl="http[s]://<rhost>[:<port>]/redfish"

        #send DELETE to the resource path specified
        rc,r,j,d=rft.rftSendRecvRequest(apiType, method, rootUrl, relPath=path)
        if(rc!=0):
            rft.printErr("raw: Error sending DELETE to resource, aborting")
            return(rc,r,False,None)


        if(rc==0):   rft.printVerbose(1," raw DELETE:",skip1=True, printV12=cmdTop)
        return(rc,r,j,d)
    
      
    def examples(self,sc,op,rft,cmdTop=False,prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
        print(" {} -r<ip> raw GET /redfish/v1/   # returns the root collection".format(rft.program))

        return(0,None,False,None)



    
'''
TODO:
1. need to handle case where no patch or post data (no -d)
2. add raw PUT
CHANGES:
0.9.2:  no longer call /redfish and /redfish/v1 before executing the raw api
'''

    





#!/usr/bin/python
# Copyright Notice:
# Copyright 2016 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfishtool/blob/master/LICENSE.md

# redfishtool:  redfishtool.py
#
# contents:
#  -CLI wrapper:  calls main() routine in ./redfishtool  package
#
#  Note: structure supports a future lib interface to the routines
#        where the libWrapper is the interface
#
from redfishtool import main
import sys

if __name__ == "__main__":
    main(sys.argv)






# Copyright Notice:
# Copyright 2016 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfishtool/blob/master/LICENSE.md

# redfishtool: SessionService.py
#
# contains SessionService related subCommands and access functions
#
# Class RfSessionServiceMain
#  - functions init, displayUsage, displayHelp, displayOperations,
#  - runOperation - SessionService command table, dispatch of operation eg get, reset
#  - SessionServiceMain - called from redfishMain, enforce legal option combinations,
#    and call runOperation to run SessionService operation (sub-sub-command)
#
# Class RfSessionServiceOperations
#  All of the SessionService sub-command operations eg: login, logout, get sessionService, etc
#  - hello - test cmd
#  - get - get the session service
#  - patch - raw subcommand to patch a sessionService, with etag support
#  - setSessionTimeout -- patches the SessionTimeout property w/ etag support
#  - Sessions - get Sessions collection, Session instance, list Sessions, get all Sessions
#  - login - Session login (post to add a new session)
#  - logout - Session logout (delete to delete a session)
#  - examples --prints some example apis
#
from   .redfishtoolTransport  import RfTransport
import requests
import json
import getopt
import re
import sys
from    .ServiceRoot import RfServiceRoot
from   urllib.parse import urljoin

class RfSessionServiceMain():
    def __init__(self):
        # operation string and remaining args
        self.operation=None
        self.args=None
        self.argnum=0
        self.nonIdCommands=None

    def displayUsage(self,rft):
        if(rft.quiet): return(0)
        print("  Usage:")
        print("   {} [OPTNS]  SessionService  <operation> [<args>]  -- perform <operation> on the SessionService  ".format(rft.program))

    def displayHelp(self,rft):
        self.displayUsage(rft)
        self.displayOperations(rft)
        print("")
   
    def displayOperations(self,rft):
        print("  <operations>:")
        print("     [get]                     -- get the sessionService object. ")
        print("     patch {A: B,C: D,...}     -- patch the sessionService w/ json-formatted {prop: value...} ")
        print("     setSessionTimeout <timeout> -- patches the SessionTimeout property w/ etag support ")
        print("     Sessions [list]           -- get the \"Sessions\" collection, or list \"Id\", username, and Url ")
        print("       Sessions [IDOPTN]       --   get the member specified by IDOPTN: -i<Id>, -m<prop>:<val>, -l<link>, -a #all")
        print("     login                     -- sessionLogin.  post to Sessions collection to create a session")
        print("                                   the user is -u<user>, password is -p<password>")
        print("     logout                    -- logout or delete the session by identified by -i<SessionId> or -l<link>")
        print("                                   where <link> is the session path returned in Location from login")
        print("     examples                  -- example commands with syntax")
        print("     hello                     -- Systems hello -- debug command")
        return(0)


    def runOperation(self,rft):
        #  instantiate SessionServiceOperations class
        op=RfSessionServiceOperations()
        
        #  dispatch table for each subcommand:   "cmdName": cmdClass.cmdFunction"
        operationTable = {
            "get":                          op.get,
            "patch":                        op.patch,
            "setSessionTimeout":            op.setSessionTimeout,
            "Sessions":                     op.getSessions,
            "login":                        op.sessionLogin,
            "logout":                       op.sessionLogout,
            "hello":                        op.hello,
            "examples":                     op.examples
        }

        rft.printVerbose(5,"SessionService:runOperation: operation: {}".format(self.operation))
        rft.printVerbose(5,"SessionService:runOperation: args:  {}".format(self.args))
            
        if self.operation in operationTable:
            rft.printVerbose(5,"SessionService:runOperation: found Oper: {} in table. executing".format(rft.subcommand))
            rc,r,j,d=operationTable[self.operation](self, op, rft, cmdTop=True)
            return(rc,r,j,d)
        
        else: # invalid operation
            rft.printErr("SessionService: Invalid operation: {}".format(self.operation))
            return(2,None,False,None)
        


    def SessionServiceMain(self,rft,cmdTop=False):
        rft.printVerbose(4,"SessionServiceMain:  subcommand: {}".format(rft.subcommand))
        
        if( rft.help ):
            self.displayHelp(rft)
            return(0,None,False,None)
        
        # we will validate usage of -P and -a in action processing
        # actually, if a non 'get' action is specified, -P and -a are just ignored :)

        args=rft.subcommandArgv[0:]
        
        #if no args, this is a getSessionService command
        if(  len(args) < 2 ):
            self.operation="get"
            self.args= None
        else:
            self.operation=args[1]
            self.args = args[1:]        # now args points to the 1st argument
            self.argnum =len(self.args)
            
        rft.printVerbose(5,"SessionService: operation={}, args={}".format(self.operation,self.args))
                          
        # now execute the operation.
        rc,r,j,d = self.runOperation(rft)
                
        if(rc !=0 ):
            rft.printVerbose(5,"SessionService: operation returned with error: rc={}".format(rc))
            return(rc,r,False,None)
        
        #else, if here, the subcommand executed without error.  Return with 0 exit code
        rft.printVerbose(5,"SessionService: operation exited OK")
        return(rc,r,j,d)


#
# contains operations related to the SessionService subCommand
#
class RfSessionServiceOperations():
    def __init__(self):
        self.SessionServicePath=None
        self.SessionServiceCollectionDict=None


    def hello(self,sc,op,rft,cmdTop=False):
        rft.printVerbose(4,"in hello")
        rft.printVerbose(4,"   subcmd:{}, operation:{}, args:{}".format(rft.subcommand,sc.operation,sc.args))
        print("hello world from SessionService")
        return(0,None,False,None)



    def get(self,sc,op,rft, cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))

        # 1st get serviceRoot
        svcRoot=RfServiceRoot()
        rc,r,j,d = svcRoot.getServiceRoot(rft)
        if( rc != 0 ):
            rft.printErr("get SessionService: Error getting service root, aborting")
            return(rc,r,False,None)

        # get the link to the SessionService
        # need to test we got good data
        if (("SessionService" in d) and ("@odata.id" in d["SessionService"])):
            sessionServiceLink=d["SessionService"]["@odata.id"]
        else:
            rft.printErr("Error:  root does not have a SessionService link")
            return(4)
        
        rft.printVerbose(4,"SessionService: get SessionService: link is: {}".format(sessionServiceLink))
      
        if cmdTop is True:   prop=rft.prop
              
        # do a GET to get the SessionService, if -P show property, else show full response
        rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=sessionServiceLink, prop=prop)

        if(rc==0):   rft.printVerbose(1," SessionService Resource:",skip1=True, printV12=cmdTop)
        return(rc,r,j,d)


    
    def patch(self,sc,op,rft,cmdTop=False, prop=None, patchData=None, r=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
        # verify we have got an argument which is the patch structure
        # its in form '{ "AssetTag": <val>, "IndicatorLed": <val> }'
        ##print("patchData: {}".format(patchData))
        if( len( sc.args) == 2):
            ##print("data:{}".format(sc.args[1]))
            try:
                patchData=json.loads(sc.args[1])
            except ValueError:
                rft.printErr("Patch: invalid Json input data:{}".format(sc.args[1]))
                return(5,None,False,None)
            ##print("patchData: {}".format(patchData))
        else:
            rft.printErr("Patch: error: invalid argument format")
            rft.printErr("     : args={}".format(sc.args))
            rft.printErr("     : expect: SessionService patch \"{ <prop>: <value> }\"")
            return(4,None,False,None)

        # read the sessionServiceLink resource
        # this is used by the generic rft.patchResource() function to see if there is an etag in the response
        # if an etag is in response hdr, then we must include the etag on the patch, so this get is required
        # note: the format of etag header for redfish is:  ETag: W/"<string>" or "<string"
        rc,r,j,d=op.get(sc,op,rft)
        if( rc != 0):
            return(rc,r,False,None)

        # now call the generic patch function to send the patch
        rc,r,j,d=rft.patchResource(rft, r, patchData)

        if(rc==0):   rft.printVerbose(1," SessionService Patch:",skip1=True, printV12=cmdTop)
        return(rc,r,j,d)


        
    def setSessionTimeout(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))

        propName="SessionTimeout"
        
        # get the SessionTimeout from args
        if(len(sc.args) < 2 ):
            rft.printErr("Error, no SessionTimeout value specified")
            rft.printErr("Syntax:  {} [options] SessionService setSessionTimeout <timeoutInt> ".format(rft.program))
            return(8,None,False,None)
        sessTimeout=int(sc.args[1],0)  # base 0 means int() will interpret 0x<num> as hex and <num> as decimal
        patchData={propName: sessTimeout}
        
        # get the resource to verify it includes AssetTag,
        # the response will also be used by Patch() to check for etag 
        rc,r,j,d=op.get(sc,op,rft)
        if( rc != 0):  return(rc,r,False,None)
        if( not propName in d ):
            rft.printErr("SessionService resource does not have a {} property.".format(propName))
            return(8,r,False,None)

        rc,r,j,d=rft.patchResource(rft, r, patchData)
        if(rc==0):
            rft.printVerbose(1," SessionService setSessionTimeout:",skip1=True, printV12=cmdTop)
            assetTag={propName: d[propName]}
            return(rc,r,j,assetTag)         
        else: return(rc,r,False,None)



    def getSessions(self,sc,op, rft, cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation: getSessions collection".format(rft.subcommand,sc.operation))

        # get the system resource
        rc,r,j,d=op.get(sc,op, rft)     
        if( rc != 0):  return(rc,r,False,None)
        
        collName="Sessions"
        # get the link to the Sessions collection
        if ((collName in d) and ("@odata.id" in d[collName])):
            sessionsLink=d[collName]["@odata.id"]
        else:
            rft.printErr("Error: SessionService resource does not have a {} link".format(collName))
            return(6,None,False,None)
        
        if cmdTop is True:   prop=rft.prop

        # check if there is a list arg for the operation
        if( sc.argnum > 1 and sc.args[1] == 'list' ):
            #get the collection
            rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=sessionsLink)
            #loop through the members and create the list sub-operation response
            rc,r,j,d=rft.listCollection(rft, r, d, prop="UserName")
            if(rc==0):
                rft.printVerbose(1," list {} Collection member info: Id, URI, Socket".format(collName,skip1=True, printV12=cmdTop))

        # else: check if no session was specified.  If not, return the collection
        elif(rft.IdLevel2OptnCount==0):
            rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=sessionsLink, prop=prop)
            if(rc==0):
                rft.printVerbose(1," {} Collection ".format(collName,skip1=True, printV12=cmdTop))

        # else:  check if the -a (all) option is set. If not, return the session specific by -i or -m or -l
        # search collection to find path using getPath2 
        elif( rft.allOptn is not True ):
            # get the Sessions collection
            rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=sessionsLink, prop=prop)
            collUrl=r.url

            # now search for 2nd level resource and return
            path2,rc,r,j,d=rft.getLevel2ResourceById(rft,r,d)
            if(rc!=0):
                return(rc,r,j,d)
            # so rc=0
            #if sysPath returned a response but we need to extract the property do it here
            if( (r is not None) and (prop is not None) ):
                rc,r,j,d=rft.getPropFromDict(rft,r,d,prop)

            # otherwise, we need to do a GET to get the session, if -P show property, else show full response
            elif( r is None ):
                rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', collUrl, relPath=path2, prop=prop)
                if(rc==0):
                    rft.printVerbose(1," {} Collection Member ".format(collName,skip1=True, printV12=cmdTop))

        # else, return ALL of the Sessions members
        else:
            rft.printVerbose(4,"getting expanded Sessions Collection")
            rc,r,j,d=rft.getAllCollectionMembers(rft, r.url, relPath=sessionsLink)
            if(rc==0):
                rft.printVerbose(1," Get ALL {} Collection Members".format(collName,skip1=True, printV12=cmdTop))
        
        return(rc,r,j,d)


    

    #SessionService -u <username> -p <passwd> login,  returns authToken and sessionId
    def sessionLogin(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))

        #note that <username> and <password> were passed-in in -u and -p options and are stores in
        # rft.user and rft.password

        # 1st get serviceRoot
        svcRoot=RfServiceRoot()
        rc,r,j,d = svcRoot.getServiceRoot(rft)
        if( rc != 0 ):
            rft.printErr("SessionService: login: Error getting service root, aborting")
            return(rc,r,False,None)

        #call rfSessionLogin in transport class to login
        rc,r,j,d=rft.rfSessionLogin(rft,cleanupOnExit=False)
        if(rc == 0):
            # return sessionId and authToken
            respData={"SessionId": rft.sessionId, "SessionLocation": rft.sessionLink, "X-Auth-Token": rft.authToken}
            rft.printVerbose(1," SessionLogin: {}".format(respData, skip1=True, printV12=cmdTop))
            return(rc,r,j,respData)
        else:
            return(rc,r,j,d)


    #SessionService -t<token> logout [-i<sessionId>|-l<sessionLink>],   returns 204 no content
    def sessionLogout(self,sc,op,rft,cmdTop=False, prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))

        if( (rft.linkLevel2 is None) and( rft.IdLevel2 is None) ):
            rft.printErr("Error, logout:  no sessionId or sessionLink specified")
            rft.printErr("Syntax:  {} [options] [SessionService] logout [-i<sessionId> | -l <sessionLink>]".format(rft.program))
            return(8,None,False,None)

        # 1st get serviceRoot
        svcRoot=RfServiceRoot()
        rc,r,j,d = svcRoot.getServiceRoot(rft)
        if( rc != 0 ):
            rft.printErr("SessionService: login: Error getting service root, aborting")
            return(rc,r,False,None)
            
        #find Sessions collection
        if( ("Links" in d) and ("Sessions" in d["Links"]) and ("@odata.id" in d["Links"]["Sessions"]) ):
            sessionsLink=rft.rootResponseDict["Links"]["Sessions"]["@odata.id"]
            #print("loginUri:{}".format(loginUri))
        else:
            rft.printErr("Error: the rootService response does not have a login link: \"Links\":\"Sessions\":{{\"@odata.id\": <uri>}")
            return(4,None,False,None)
            
        # get the Sessions collection
        rc,r,j,d=rft.rftSendRecvRequest(rft.AUTHENTICATED_API, 'GET', r.url, relPath=sessionsLink, prop=prop)
        if(rc != 0):
            rft.printErr("Error: logout: cant read sessions collection")
            return(6,None,False,None)
        # now search for 2nd level resource and return
        path2,rc,r,j,d=rft.getLevel2ResourceById(rft,r,d)
        if(rc!=0):
            rft.printErr("Error: the specified sessionId or sessionLink was not a valid Sessions collection entry")
            return(rc,r,j,d)

        #path2 is the path the the session we want to delete
        rc,r,j,d=rft.rfSessionDelete(rft,sessionLink=path2)
        if(rc!=0):
            rft.printErr("Error: Logout: the session delete failed")
            return(rc,r,j,d)
        else:
            rft.printVerbose(1," SessionLogout successful".format(skip1=True, printV12=cmdTop))
            return(rc,r,j,d)


    
      
    def examples(self,sc,op,rft,cmdTop=False,prop=None):
        rft.printVerbose(4,"{}:{}: in operation".format(rft.subcommand,sc.operation))
        print(" {} -r<ip> SessionService                   # gets the sessionService".format(rft.program))
        print(" {} -r<ip> SessionService setSessionTimeout <timeout> # sets the session timeout property".format(rft.program))
        print(" {} -r<ip> SessionService Sessions             # gets Sessions collection".format(rft.program))
        print(" {} -r<ip> SessionService Sessions -l<sessUrl> # gets the session at URI=<sessUrl".format(rft.program))
        print(" {} -r<ip> SessionService Sessions -i<sessId>  # gets the session with session Id <sessId>".format(rft.program))
        print(" {} -r<ip> SessionService patch {{A: B,C: D,...}} # patch the json-formatted {{prop: value...}} data to the sessionService object".format(rft.program))
        print(" {} -r<ip> SessionService login <usernm> <passwd> # login (create session)".format(rft.program))
        print(" {} -r<ip> SessionService logout <sessionId>      # logout (delete session <sessId>".format(rft.program))
        return(0,None,False,None)


    
'''
TODO:
1. e

'''

    





from setuptools import setup
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(name='redfishtool',
      version='1.1.0',
      description='Redfishtool package and command-line client',
      long_description=long_description,
      long_description_content_type='text/markdown',
      author='DMTF, https://www.dmtf.org/standards/feedback',
      license='BSD 3-clause "New" or "Revised License"',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: BSD License',
          'Programming Language :: Python :: 3.4',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Communications'
      ],
      keywords='Redfish',
      url='https://github.com/DMTF/Redfishtool',
      download_url='https://github.com/DMTF/Redfishtool/archive/1.0.5.tar.gz',
      packages=['redfishtool'],
      scripts=['scripts/redfishtool'],
      install_requires=['requests']
      )



# Copyright Notice:
# Copyright 2016 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfishtool/blob/master/LICENSE.md

# redfishtool:  redfishtool.py  Main
#
# contains:
#  - functions called for usage:
#     -- displayUsage, displayOptions, listSubcommands, displayHelp
#  - Main routine, with argc/argv option parsing, and valid argument/option checking
#    after parsing options, runs cmd with runSubCmd function, outputs result, cleansup sessions created
#  - runSubCmd -- Main subcommand command table and subcommand execution
#  - hello subcommand for testing
#  - help  subcommand
#  - about subcommand for info about this version of redfishtool
#
import sys
import getopt
import re
import json
from .redfishtoolTransport   import RfTransport
from .ServiceRoot import RfServiceRoot
from .Systems import RfSystemsMain
from .Chassis import RfChassisMain
from .Managers import RfManagersMain
from .SessionService import RfSessionServiceMain
from .AccountService import RfAccountServiceMain
from .raw import RfRawMain

def displayUsage(rft,*argv,**kwargs):
        rft.printErr("  Usage:",noprog=True)
        rft.printErr("   {} [OPTIONS]  <SubCommand> <operation> [<args>]... ",prepend="  ")
        rft.printErr("   {} [OPTIONS]  hmraw  <method> <hmUrl> [<data>]",     prepend="  ")
              
def displayOptions(rft):
        print("")
        print("  Common OPTIONS:")
        print("   -V,          --version           -- show {} version, and exit".format(rft.program))
        print("   -h,          --help              -- show Usage, Options, and list of subCommands, and exit".format(rft.program))
        print("   -v,          --verbose           -- verbose level, can repeat up to 5 times for more verbose output")
        print("                              -v(header), -vv(+addl info), -vvv(Request trace), -vvvv(+subCmd dbg), -vvvvv(max dbg)")
        print("   -s,          --status            -- status level, can repeat up to 5 times for more status output")
        print("                               -s(http_status), ")
        print("                               -ss(+r.url, +r.elapsed executionTime ), ")
        print("                               -sss(+request hdrs,data,authType, +response status_code, +response executionTime, ")
        print("                                    +login auth token/sessId/sessUri)")
        print("                               -ssss(+response headers), -sssss(+response data")
        print("   -u <user>,   --user=<usernm>     -- username used for remote redfish authentication")
        print("   -p <passwd>, --password=<passwd> -- password used for remote redfish authentication")
        print("   -r <rhost>,  --rhost=<rhost>     -- remote redfish service hostname or IP:port")
        print("   -t <token>,  --token=<token>     -- redfish auth session token-for sessions across multiple calls")
        print("   -q,          --quiet             -- quiet mode--suppress error, warning, and diagnostic messages")
        print("   -c <cfgFile>,--config=<cfgFile>  -- read options (including credentials) from file <cfgFile>")
        print("   -T <timeout>,--Timeout=<timeout> -- timeout in seconds for each http request.  Default=10")
        print("")
        print("   -P <property>, --Prop=<property> -- return only the specified property. Applies only to all \"get\" operations")
        print("   -E, --Entries                    -- Fetch the Logs entries. Applies to Logs sub-command of Systems, Chassis and Managers")
        print("")
        print("  Options used by \"raw\" subcommand:")
        print("   -d <data>    --data=<data>       -- the http request \"data\" to send on PATCH,POST,or PUT requests")
        print("")
        print("  Options to specify top-level collection members: eg: Systems -I <sysId>")
        print("   -I <Id>, --Id=<Id>               -- Use <Id> to specify the collection member")
        print("   -M <prop>:<val> --Match=<prop>:<val>-- Use <prop>=<val> search to find the collection member")
        print("   -F,  --First                     -- Use the 1st link returned in the collection or 1st \"matching\" link if used with -M")
        print("   -1,  --One                       -- Use the single link returned in the collection. Return error if more than one member exists")
        print("   -a,  --all                       -- Returns all members if the operation is a Get on a top-level collection like Systems")
        print("   -L <Link>,  --Link=<Link>        -- Use <Link> (eg /redfish/v1/Systems/1) to reference the collection member. ")
        print("                                    --   If <Link> is not one of the links in the collection, and error is returned.")
        print("  Options to specify 2nd-level collection members: eg: Systems -I<sysId> Processors -i<procId>")
        print("   -i <id>, --id=<id>               -- use <id> to specify the 2nd-level collection member")
        print("   -m <prop>:<val> --match=<prop>:val>--use <prop>=<val> search of 2nd-level collection to specify member")
        print("   -l <link>  --link=<link>         -- Use <link> (eg /redfish/v1/SYstems/1/Processors/1) to reference a 2nd level resource")
        print("                                    --   A -I|M|F|1|L option is still required to specify the link to the top-lvl collection")
        print("   -a,  --all                       -- Returns all members of the 2nd level collection if the operation is a Get on the ") 
        print("                                    --   2nd level collection (eg Processors). -I|M|F|1|L still specifies the top-lvl collection.")
        print("")
        print("  Additional OPTIONS:")
        print("   -W <num>:<connTimeout>,          -- Send up to <num> {GET /redfish} requests with <connTimeout> TCP connection timeout")
        print("         --Wait=<num>:<ConnTimeout> --   before sending subcommand to rhost.  Default is -W 1:3")
        print("   -A <Authn>,   --Auth <Authn>     -- Authentication type to use:  Authn={None|Basic|Session}  Default is Basic")
        print("   -S <Secure>,  --Secure=<Secure>  -- When to use https: (Note: doesn't stop rhost from redirect http to https)")
        print("                                       <Secure>={Always | IfSendingCredentials | IfLoginOrAuthenticatedApi(default) }")
        print("   -R <ver>,  --RedfishVersion=<ver>-- The Major Redfish Protocol version to use: ver={v1(dflt), v<n>, Latest}")
        print("   -C         --CheckRedfishVersion -- tells Redfishtool to execute GET /redfish to verify that the rhost supports")
        print("                                       the specified redfish protocol version before executing a sub-command. ")
        print("                                       The -C flag is auto-set if the -R Latest or -W ... options are selected")
        print("   -H <hdrs>, --Headers=<hdrs>      -- Specify the request header list--overrides defaults. Format \"{ A:B, C:D...}\" ")
        print("   -D <flag>,  --Debug=<flag>       -- Flag for dev debug. <flag> is a 32-bit uint: 0x<hex> or <dec> format")
        print("")
        
def listSubcommands(rft):
        # print a list of Main subcommands in order to show them in lc subcommand
        print("  Subcommands:")
        print("     hello                 -- redfishtool hello world subcommand for dev testing")
        print("     about                 -- display version and other information about this version of {}".format(rft.program))
        print("     versions              -- get redfishProtocol versions supported by rhost: GET ^/redfish")
        print("     root   |  serviceRoot -- get serviceRoot resouce: GET ^/redfish/v1/")
        print("     Systems               -- operations on Computer Systems in the /Systems collection ")
        print("     Chassis               -- operations on Chassis in the /Chassis collection")
        print("     Managers              -- operations on Managers in the /Managers collection")
        print("     AccountService        -- operations on AccountService including user administration")
        print("     SessionService        -- operations on SessionService including Session login/logout")
        print("     odata                 -- get the Odata Service document: GET ^/redfish/v1/odata")
        print("     metadata              -- get the CSDL metadata document: GET ^/redfish/v1/$metadata")
        print("     raw                   -- subcommand to execute raw http methods(GET,PATCH,POST...) and URIs")

        return(0)

def displayHelp(rft):
        displayUsage(rft,file=sys.stdout)
        displayOptions(rft)
        listSubcommands(rft)
        print("")
        print("  For Subcommand usage, options, operations, help:")
        print("     {} <SubCommand> -h  -- usage and options for specific subCommand".format(rft.program))
        print("")

def main(argv):
    #instantiate transport object which initializes default options
    rft=RfTransport()

    try:
        opts, args = getopt.gnu_getopt(argv[1:],"Vhvsqu:p:r:t:c:T:P:d:EI:M:F1L:i:m:l:aW:A:S:R:H:D:C",
                        ["Version", "help", "verbose", "status", "quiet", 
                         "user=", "password=", "rhost=", "token=", "config=", "Timeout=",
                         "Prop=", "data=", "Entries", "Id=", "Match=", "First", "One", "Link=",
                         "id=", "match=", "link", "all",
                         "Wait=", "Auth=","Secure=", "RedfishVersion=", "Headers=", "Debug=",
                         "CheckRedfishVersion"  ])
    except getopt.GetoptError:
        rft.printErr("Error parsing options")
        displayUsage(rft)
        sys.exit(1)
        
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            rft.help=True
        elif opt in ("-V", "--Version", "--version"):
            print("{} Version: {}".format(rft.program, rft.version))
            sys.exit(0)
        elif opt in ("-v", "--verbose"):
            rft.verbose = min((rft.verbose+1), 5)
        elif opt in ("-s", "--status"):
            rft.status = min((rft.status+1), 5)
        elif opt in ("-q", "--quiet"):
            rft.quiet=True
        elif opt in ("-u", "--user"):
            rft.user=arg
        elif opt in ("-p", "--password"):
            rft.password=arg
        elif opt in ("-r", "--rhost"):
            rft.rhost=arg
        elif opt in ("-t", "--token"):          #Redfish Session authentication token
            rft.token=arg
            rft.authToken=rft.token
        elif opt in ("-c", "--config"):         #read additional options from a file
            rft.configFile=arg
            try:
                with open(arg, 'r') as f:
                    configdata = json.load(f)
                if 'user' in configdata and 'password' in configdata:
                    rft.user = configdata['user']
                    rft.password = configdata['password']
                    rft.printVerbose(1,"Main: get pw from configfile, usr:{}, pw:{}".format(rft.user, rft.password))
                else:
                    rft.printErr("Invalid --config= filedata: {}".format(configdata))
            except IOError:
                rft.printErr("Invalid: Failed to read configfile")

        elif opt in ("-T", "--Timeout"):        #Specify http timeout in  seconds
            timePattern="^([1-9][0-9]*)$"
            timeMatch=re.search(timePattern,arg)
            if( timeMatch ):
                rft.timeout=int(arg)
            else:
                rft.printErr("Invalid -Timeout value: {}".format(arg))
                rft.printErr("     Expect: -T <timeout> where <timeout> is a decimal int",noprog=True)
                sys.exit(1)                
        # options that reference a specific property for Get operations
        elif opt in ("-P", "--Prop"):       
            rft.prop=arg
            rft.gotPropOptn=True
        # options used by raw subcommand to specify http method (-X, --request) and request data (-d, --data)
        elif opt in ("-d", "--data"):
            rft.requestData=arg
        # options related to the Logs sub-command of Systems, Chassis and Managers
        elif opt in ("-E", "--Entries"):
            rft.gotEntriesOptn = True
        # options to find the Id or member of a collection in a subcommand
        elif opt in ("-I", "--Id"):
            rft.Id=arg
            rft.gotIdOptn=True
            rft.IdOptnCount+=1
        elif opt in ("-M", "--Match"):  
            # arg is of the form: "<prop>:<value>"
            pair = arg.split(':', 1)
            if len(pair) == 2:
                rft.matchProp = pair[0]
                rft.matchValue = pair[1]
                rft.IdOptnCount += 1
                rft.gotMatchOptn = True
            else:
                rft.printErr("Invalid --Match= option format: {}".format(arg))
                rft.printErr("     Expect --Match=<prop>:<value> Ex -M AssetTag:5555, -Match=AssetTag:5555",noprog=True)
                sys.exit(1)
        elif opt in ("-F", "--First"):        
            rft.firstOptn=True
            rft.IdOptnCount+=1
        elif opt in ("-1", "--One"):
            rft.oneOptn=True
            rft.IdOptnCount+=1
        elif opt in ("-L", "--Link"):       
            rft.Link=arg
            rft.gotIdOptn=True
            rft.IdOptnCount+=1
        elif opt in ("-i", "--id"):
            rft.IdLevel2=arg
            rft.gotIdLevel2Optn=True
            rft.IdLevel2OptnCount+=1
        elif opt in ("-m", "--match"):
            # arg is of the form: "<prop>:<value>"
            pair = arg.split(':', 1)
            if len(pair) == 2:
                rft.matchLevel2Prop = pair[0]
                rft.matchLevel2Value = pair[1]
                rft.IdLevel2OptnCount += 1
                rft.gotMatchLevel2Optn = True
            else:
                rft.printErr("Invalid level2 --match= option format: {}".format(arg))
                rft.printErr("     Expect --match=<prop>:<value> Ex -m ProcessorType:CPU, -match=ProcessorType:CPU",noprog=True)
                sys.exit(1)
        elif opt in ("-l", "--link"):       
            rft.linkLevel2=arg
            rft.gotIdLevel2Optn=True
            rft.IdLevel2OptnCount+=1
        elif opt in ("-a", "--all"):        
            rft.allOptn=True        # additional options
            rft.IdLevel2OptnCount+=1
        elif opt in ("-W", "--Wait"):           #specify how long to ping rhost before sending http requests 0=no ping
            # arg is of the form "<num>:<time>
            rft.checkProtocolVer=True           # if -W ..., set flag to force running GET /redfish first
            waitPattern="^([1-9][0-9]*):([1-9][0-9]*)$"
            waitMatch=re.search(waitPattern,arg)
            if( waitMatch ):
                rft.waitNum=int(waitMatch.group(1))
                rft.waitTime=int(waitMatch.group(2))
                print("num: {},  time: {}".format(rft.waitNum, rft.waitTime))
            else:
                rft.printErr("Invalid --Wait= option format: {}".format(arg))
                rft.printErr("     Expect --Wait=<num>:<time>  where num and time are decimals. Ex -W 10:5",noprog=True)
                sys.exit(1)
        elif opt in ("-A", "--Auth"):           # specify authentication type
            rft.auth=arg
            if not rft.auth in rft.authValidValues:
                rft.printErr("Invalid --Auth option: {}".format(rft.auth))
                rft.printErr("   Valid values: {}".format(rft.authValidValues),noprog=True)
                sys.exit(1)
        elif opt in ("-S", "--Secure"):         #Specify when to use HTTPS
            rft.secure=arg
            if not rft.secure in rft.secureValidValues:
                rft.printErr("Invalid --Secure option: {}".format(rft.secure))
                rft.printErr("     Valid values: {}".format(rft.secureValidValues),noprog=True)
                sys.exit(1)
        elif opt in ("-R", "--RedfishVersion"):     #specify redfish protocol version to use
            rft.protocolVer=arg
            if(rft.protocolVer=="Latest"):
                rft.checkProtocolVer=True           # if -R Latest, set flag to force running GET /redfish first
        elif opt in ("-H", "--Headers"):     #specify request headers -- overrides defaults
            try:
                rft.headers=json.loads(arg)
            except ValueError:
                rft.printErr("Invalid -H arg: invalid Json data format: {}".format(arg))
                sys.exit(1)
        elif opt in ("-D", "--Debug"):  # --Debug=0x3334   or --Debug=233  regex:(^(0[xX]([0-9a-fA-F]){1,8}$))|(^0$)|(^([1-9][0-9]*)$)
            dbgPattern="(^(0[xX]([0-9a-fA-F]){1,8}$))|(^0$)|(^([1-9][0-9]*)$)"   # --Flag=0x<hex> or <decimal>:  0x3334 or 33...
            dbgMatch=re.search(dbgPattern,arg)
            if( dbgMatch ):
                rft.dgbFlag=int(arg,0)
                rft.printVerbose(4,"Main: Flag=0x{:08x}".format(rft.dbgFlag))
            else:
                rft.printErr("Invalid -Flag value: {}".format(arg))
                rft.printErr("     Expect --Flag=<flag> where <flag> is a decimal int",noprog=True)
                sys.exit(1)
        elif opt in ("-C", "--CheckRedfishVersion"):
            rft.checkProtocolVer=True
        else:
            rft.printErr("Error: Unsupported option: {}".format(opt))
            displayUsage(rft)
            sys.exit(1)

    # if no subcommand, check if -h --help option was entered, and print help if so
    # if a subcommand was entered, then the subcommand will print the help
    if( rft.help ):
        if( not args ):
            displayHelp(rft)
            sys.exit(0)

    # if -t <token>  option was specified, the --Auth=<auth> must be Session (which is the default)
    if (rft.token is not None) and (rft.auth != "Session"):
        rft.printErr("Invalid mix of --Auth and --token options")
        rft.printErr("  if --token=<authToken> is specified, --Auth must be Session", noprog=True)
        sys.exit(1)

    # check for invalid option combinations
    if( rft.IdOptnCount > 1 ):
        if( not (rft.firstOptn and rft.gotMatchOptn )  ):
            rft.printErr("Syntax error: invalid combination of -I,-M,-1,-F options.")
            rft.printErr("    Valid combinations: -I  |  -M[F]  | -1  | -F ",noprog=True)
            displayUsage(rft)
            sys.exit(1)

    # if -I <Id>, convert to -M <prop>:<val>
    if( rft.gotIdOptn is True ):
        rft.matchProp="Id"
        rft.matchValue=rft.Id
        rft.gotMatchOptn=True
        rft.firstOptn=True

        
    # if -i <id>, convert to -m <prop>:<val>
    if( rft.gotIdLevel2Optn is True ):
        rft.matchLevel2Prop="Id"
        rft.matchLevel2Value=rft.IdLevel2
        rft.gotMatchLevel2Optn=True
        
    # check for invalid Level-2 collection member reference options
    # there are 3 ways to specify the level-2 colltion member:  -i<id>  | -m<prop>:<val>   |  -l<link> | -a
    # a command should only include one of these.  each time one if found during option processing, IdLevel2OptnCount is incremented
    if( rft.IdLevel2OptnCount > 1):
        rft.printErr("Syntax error: invalid mix of options -i,-m,-a used to specify a 2nd-level collection member.")
        rft.printErr("    Valid combinations: -i  |  -m  | -l | -a ",noprog=True)
        displayUsage(rft)
        sys.exit(1)    

    # -P, -a will be validated against the operation in the Subcommand processing
    # whether at least one -I|-M|-1|-F|-L or one -i|-m|-l|-a is required is validated in subcommand processing based on operation

    #after parsing options (GNU style) args should now be a list of arguments starting with the subcommand
    #if no subcommand at this point, it is a syntax error.
    #otherwise, save the subcommand and subcommand argv array for the subcommand to parse
    if( not args ):
        rft.printErr("Syntax error. No subcommand specified.")
        displayUsage(rft)
        sys.exit(1)
    else:
        rft.subcommand=args[0]
        rft.subcommandArgv=list(args)

    rft.printVerbose(5,"Main: subcmd: {}, subCmdArgs:{}".format(rft.subcommand,rft.subcommandArgv))
    rft.printVerbose(5,"Main: verbose={}, status={}, user={}, password={}, rhost={}".format(rft.verbose, rft.status,
                                                        rft.user,rft.password,rft.rhost))
    rft.printVerbose(5,"Main: token={}, RedfishVersion={}, Auth={}, Timeout={}".format(rft.token,
                                                        rft.protocolVer, rft.auth, rft.timeout))
    rft.printVerbose(5,"Main: prop={}, Id={}, Match={}:{}, First={}, -1={}, Link={}".format( rft.prop,
                        rft.Id, rft.matchProp,rft.matchValue, rft.firstOptn, rft.oneOptn, rft.Link))
    rft.printVerbose(5,"Main: gotIdOptn={}, IdOptnCount={}, gotPropOptn={}, gotMatchOptn={}, gotEntriesOptn={}".format(
                            rft.gotIdOptn, rft.IdOptnCount, rft.gotPropOptn, rft.gotMatchOptn, rft.gotEntriesOptn))
    rft.printVerbose(5,"Main: 2nd-Level Collection Member reference options: -i<id>={}, -m<match>={}:{}, -l<link>={} -all={}".format(
                            rft.IdLevel2, rft.matchLevel2Prop, rft.matchLevel2Value, rft.linkLevel2, rft.allOptn))
    rft.printVerbose(5,"Main: 2nd-level Collection Member parsing: gotIdLevel2Optn={}, gotMatchLevel2Optn={}, IdLevel2OptnCount={}".format(
                            rft.gotIdLevel2Optn, rft.gotMatchLevel2Optn, rft.IdLevel2OptnCount ))
    rft.printVerbose(5,"Main: configFile={}, Secure={}, waitNum:waitTime={}:{}, Degug={:08x}".format(
                                        rft.configFile,rft.secure, rft.waitNum,rft.waitTime,rft.dbgFlag))
    rft.printVerbose(5,"Main: Headers={}".format(rft.headers))

    rft.printVerbose(5,"Main: options parsed.  Now lookup subcommand and execute it")

    # instansiate the SubCmd object, and run the specified subcommand
    #rfCmds=RfSubCmds()   
    #rc=rfCmds.runSubCmd(rft)
    rc,r,j,d=runSubCmd(rft)
    if(rc !=0 ):
            rft.printVerbose(5,"#DB4:Main: subcommand returned with error: rc={}".format(rc))
            rft.printVerbose(1,"Main: Error: rc={}".format(rc))
            if r is not None:
                rft.printVerbose(5,"   Response status code:{}".format(r.status_code))
                rft.printVerbose(5,"   Response headers: {}".format(r.headers))
            #cleanup any sessions we opened
            rft.rfCleanup(rft)
            sys.exit(rc)

    rft.printVerbose(5,"Main: subcommand exited OK.")
    if( r is not None ):
        rft.printVerbose(5,"    Status code:{}".format(r.status_code))
        rft.printStatus(1,r=r)
        rft.printStatus(2,r=r)
        
    # print out result here.
    if( j is True and d is not None):
        output=json.dumps(d,indent=4)
        print(output)
    elif( j is False and d is not None):
        output=r.text
        print(output)
    else:
        pass

    #cleanup any sessions we opened on the remote service
    rft.rfCleanup(rft)

    rft.printVerbose(5,"Main: Done")
    #print("headers:{}".format(r.headers))
        
    sys.exit(0)


# enter cmdClasses in other files here:


# lookup the function for the subcommand and execute it
# pass the transport class that includes argv/argc array and Main options
# returns tuple: rc(0=ok, >0=err), r (from request), dataType(0=none,1=dict,2=text) ,d (data)
# rc, r, dt, d=runSubCmd(rft)

def runSubCmd(rft):
        # instantiate all cmdClasses (note: helloCmd and listSubcommands are in this class)
        root=RfServiceRoot()
        systems=RfSystemsMain()
        chassis=RfChassisMain()
        managers=RfManagersMain()
        sessionService=RfSessionServiceMain()
        accountService=RfAccountServiceMain()
        raw=RfRawMain()

        #  dispatch table for each subcommand:   "cmdName": cmdClass.cmdFunction"
        subCmdTable = {
            "help":             helpSubcmd,
            "about":            aboutSubcmd,
            "versions":         rft.getVersions,
            "serviceRoot":      root.getServiceRoot,
            "root":             root.getServiceRoot, #alias for serviceRoot
            "odata":            root.getOdataServiceDocument,
            "metadata":         root.getOdataMetadataDocument,
            "Systems":          systems.SystemsMain,
            "Chassis":          chassis.ChassisMain,
            "Managers":         managers.ManagersMain,
            "AccountService":   accountService.AccountServiceMain,
            "SessionService":   sessionService.SessionServiceMain,
            "raw":              raw.RawMain,
            "hello":              helloSubcmd
        }

        rft.printVerbose(5,"runSubCmd: subcmd: {}".format(rft.subcommand))
        rft.printVerbose(5,"runSubCmd: argvs:  {}".format(rft.subcommandArgv))
            
        if rft.subcommand in subCmdTable:
            rft.printVerbose(5,"runSubCmd: found SubCmd: {} in table. executing".format(rft.subcommand))
            rc,r,j,d=subCmdTable[rft.subcommand](rft,cmdTop=True)
            return(rc,r,j,d)
        
        else: # invalit subcmd
            rft.printErr("Invalid SubCommand: {}".format(rft.subcommand))
            return(1,None,False,None)

    
def helloSubcmd(rft, cmdTop=False):
        rft.printVerbose(5,"Main: in hello subcommand")
        if(rft.help):
            print("# {} [OPTIONS] hello  ---prints hello world message for debug".format(rft.program))
            return(0,None,False,None)
        print("#\n# Hello World \n#")
        return(0,None,False,None)

def helpSubcmd(rft, cmdTop=False):
        rft.printVerbose(5,"Main: in help subcommand")
        displayHelp(rft)
        return(0,None,False,None)

def aboutSubcmd(rft, cmdTop=False):
        rft.printVerbose(5,"Main: in about subcommand")
        if(rft.help):
            print("# {} [OPTIONS] about  ---prints information about {}".format(rft.program,rft.program))
            return(0,None,False,None)
        print("#")
        print("# {} {}:".format(rft.program,rft.subcommand))
        print("#     Version: {}".format(rft.version))
        print("#     Supports Redfish Protocol Versions:  {}".format(rft.supportedVersions))
        print("#     Release date:  {}".format(rft.releaseDate))
        print("#     Download from: {}".format(rft.downloadFrom))
        print("#")
        return(0,None,False,None)


if __name__ == "__main__":
    main(sys.argv)


'''
TODO
1. implement -c cfgfile
2. implement -d <data>   -d @<file>  for raw subcommand POST or PATCH data
3.


'''




# Copyright Notice:
# Copyright 2016 DMTF. All rights reserved.
# License: BSD 3-Clause License. For full text see link: https://github.com/DMTF/Redfishtool/blob/master/LICENSE.md

# redfishtool: ServiceRoot.py
#
# contains serviceRoot related subCommands and access functions
# Class RfServiceRoot
#  - getServiceRoot   GET /redfish/v1
#  - getOdataServiceDocument    GET /redfish/v1/odata
#  - getOdataMetadataDocument   GET /redfish/v1/$metadata
#
import requests
import json
from urllib.parse import urljoin, urlparse, urlunparse

class RfServiceRoot:
    def __init__(self):
        # option parsing variables
        self.serviceRootDict=None
        self.url=None
        
    def getServiceRoot(self,rft, cmdTop=False, prop=None):
        rft.printVerbose(4,"ServiceRoot: in getServiceRoot")
        
        if(rft.help):
            print(" {} {} -r<rhost> [-vh]   -- get serviceRoot resource".format(rft.program,rft.subcommand))
            return(0,None,False,None)
        
        # execute GET /redfish  to negotiate protocol version and get root path
        # rootPath is stored in rft.rootPath
        rc,r,j,d=rft.getVersionsAndSetRootPath(rft)
        if( rc!=0): return(rc,r,j,d)
        #
        if cmdTop is True:  prop=rft.prop

        # get root service.  if -P prop, showproperty
        rc,r,j,d=rft.rftSendRecvRequest(rft.UNAUTHENTICATED_API, 'GET', rft.rootUri, relPath=rft.rootPath,prop=prop)

        #save the rootService response.  The transport may need it later to get link to session and login
        rft.rootResponseDict=d
        
        if(rc==0 and cmdTop is True):
            rft.printVerbose(1," Service Root:",skip1=True, printV12=cmdTop)
        return(rc,r,j,d)

        
    def getOdataServiceDocument(self,rft,cmdTop=False):
        rft.printVerbose(4,"ServiceRoot: in getOdataServiceDocument")
        
        if(rft.help):
            print(" {} {} -r<rhost> [-vh]   -- get the Odata Service Document".format(rft.program,rft.subcommand))
            return(0,None,False,None)
        
        #get root path, scheme, and create URL
        rc,r,j,d=rft.getVersionsAndSetRootPath(rft)
        if( rc!=0): return(rc,r,j,d)

        #calculate relative path = rootPath / odata
        rpath=urljoin(rft.rootPath, "odata")
        
        rc,r,j,d=rft.rftSendRecvRequest(rft.UNAUTHENTICATED_API, 'GET', rft.rootUri, relPath=rpath)
        if(rc==0 ):
            rft.printVerbose(1," Odata Service Document:",skip1=True, printV12=cmdTop)
        return(rc,r,j,d)


        
    def getOdataMetadataDocument(self,rft,cmdTop=False):
        rft.printVerbose(4,"ServiceRoot: in getOdataMetadataDocument")
        
        if(rft.help):
            print(" {} {} -r<rhost> [-vh]   -- get the CSDL metadata document".format(rft.program,rft.subcommand))
            return(0,None,False,None)
        
        #get root path, scheme, and create URL
        rc,r,j,d=rft.getVersionsAndSetRootPath(rft)
        if( rc!=0): return(rc,r,j,d)

        #calculate relative path = rootPath / $metadata
        rpath=urljoin(rft.rootPath, "$metadata")

        # set content-type to xml  (dflt is application/json)
        hdrs = {"Accept": "application/xml", "OData-Version": "4.0" }


        rc,r,j,d=rft.rftSendRecvRequest(rft.UNAUTHENTICATED_API, 'GET', rft.rootUri, relPath=rpath, jsonData=False, headersInput=hdrs )
        if(rc==0):
            rft.printVerbose(1," Odata Metadata Document:",skip1=True,printV12=cmdTop)
        return(rc,r,j,d)

'''
TODO:
1.


'''
